from annolid.gui.widgets.extract_frame_dialog import ExtractFrameDialog
from annolid.gui.widgets.covert_coco_dialog import ConvertCOODialog
from annolid.gui.widgets.train_model_dialog import TrainModelDialog